# Copyright 2022 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from EvoMSA.tests.test_base import TWEETS
from microtc.utils import tweet_iterator
import numpy as np


def test_EvoDAG_decision_function():
    from EvoMSA.evodag import EvoDAG, TextRepresentations
    D = list(tweet_iterator(TWEETS))
    class _EvoDAG(TextRepresentations):
        def estimator(self):
            return EvoDAG(n_estimators=2, 
                          max_training_size=100)    
    evodag = _EvoDAG(keyword=False, emoji=False,
                     decision_function='decision_function').fit(D)
    hy = evodag.decision_function(D)    
    assert hy.shape[0] == 1000 and hy.shape[1] == 4


def test_EvoDAG_predict():
    from EvoMSA.evodag import EvoDAG, TextRepresentations
    D = list(tweet_iterator(TWEETS))
    class _EvoDAG(TextRepresentations):
        def estimator(self):
            return EvoDAG(n_estimators=2, 
                          max_training_size=100)    
    evodag = _EvoDAG(keyword=False, emoji=False).fit(D)
    hy = evodag.predict(D)
    assert (hy == [x['klass'] for x in D]).mean() > 0.25
    

def test_BoW_train_predict_decision_function():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    bow = BoW(lang='es')
    hy = bow.train_predict_decision_function(D)
    assert isinstance(hy, np.ndarray)
    assert hy.shape[0] == len(D)
    bow = BoW(lang='es')


def test_BoW_predict():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    bow = BoW(lang='es').fit(D)
    hy = bow.predict(D)
    assert hy.shape[0] == len(D)


def test_BoW_decision_function():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    bow = BoW(lang='es').fit(D)
    hy = bow.decision_function(D)
    assert hy.shape[0] == len(D)
    bow = BoW(lang='es').fit([x for x in D if x['klass'] in ['P', 'N']])
    hy = bow.decision_function(D)
    assert hy.shape[0] == len(D)


def test_BoW_key():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    O = BoW(lang='es').transform(D)    
    X = [dict(klass=x['klass'], premise=x['text']) for x in D]    
    bow = BoW(lang='es', key='premise')
    assert abs(bow.transform(X) - O).sum() == 0
    X = [dict(klass=x['klass'], premise=x['text'], conclusion=x['text']) for x in D]
    bow = BoW(lang='es', key=['premise', 'conclusion'])
    assert abs(bow.transform(X) - O * 2).sum() == 0


def test_TextRepresentations_transform():
    from EvoMSA.evodag import TextRepresentations
    D = list(tweet_iterator(TWEETS))
    text_repr = TextRepresentations(lang='es', keyword=False, emoji=False)
    X = text_repr.transform(D)
    assert X.shape[0] == len(D)
    assert len(text_repr.text_representations) == X.shape[1]


def test_TextRepresentations_fit():
    from EvoMSA.evodag import TextRepresentations
    D = list(tweet_iterator(TWEETS))
    text_repr = TextRepresentations(lang='es', keyword=False, emoji=False).fit(D)
    text_repr.predict(['Buen dia'])


def test_TextRepresentations_key():
    from EvoMSA.evodag import TextRepresentations
    D = list(tweet_iterator(TWEETS))
    O = TextRepresentations(lang='es', unit_vector=False,
                            keyword=False, emoji=False).transform(D)    
    X = [dict(klass=x['klass'], premise=x['text'], conclusion=x['text']) for x in D]
    bow = TextRepresentations(lang='es', unit_vector=False,
                              keyword=False, emoji=False, key=['premise', 'conclusion'])
    assert abs(bow.transform(X) - O * 2).sum() == 0    


def test_StackGeneralization():
    from EvoMSA.evodag import StackGeneralization, BoW, TextRepresentations
    D = list(tweet_iterator(TWEETS))
    text_repr = StackGeneralization(lang='es',
                                    decision_function_models=[BoW(lang='es')],
                                    transform_models=[TextRepresentations(lang='es', keyword=False, emoji=False)]).fit(D)
    assert text_repr.predict(['Buen dia'])[0] == 'P'


def test_StackGeneralization_train_predict_decision_function():
    from EvoMSA.evodag import StackGeneralization, BoW, TextRepresentations
    D = list(tweet_iterator(TWEETS))
    text_repr = StackGeneralization(lang='es',
                                    decision_function_models=[BoW(lang='es')],
                                    transform_models=[TextRepresentations(lang='es', keyword=False, emoji=False)])
    hy = text_repr.train_predict_decision_function(D)
    assert hy.shape[0] == len(D)
    D1 = [x for x in D if x['klass'] in ['P', 'N']]
    hy = text_repr.train_predict_decision_function(D1)
    assert hy.shape[1] == 2


def test_TextRepresentations_tr_setter():
    from EvoMSA.evodag import TextRepresentations
    D = list(tweet_iterator(TWEETS))
    text_repr = TextRepresentations(lang='es', keyword=False, emoji=False)
    tr = text_repr.text_representations
    text_repr.text_representations = tr[:3]
    assert text_repr.transform(['Buen dia']).shape[1] == 3

def test_BoW_setter():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    bow = BoW(lang='es')
    bow.bow = True
    assert bow._bow


def test_TextRepresentations_names():
    from EvoMSA.evodag import TextRepresentations
    text_repr = TextRepresentations(lang='es', keyword=False, emoji=False)
    X = text_repr.transform(['buenos dias'])
    assert X.shape[1] == len(text_repr.names)


def test_TextRepresentations_select():
    from EvoMSA.evodag import TextRepresentations
    text_repr = TextRepresentations(lang='es', keyword=False, emoji=False)
    text_repr.select([1, 10, 11])
    X = text_repr.transform(['buenos dias'])
    assert X.shape[1] == len(text_repr.names)


def test_BoW_pretrain_False():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    
    bow = BoW(lang='es', pretrain=False,
              b4msa_kwargs=dict(max_dimension=True,
                                token_max_filter=2**10)).fit(D)
    X = bow.transform(D)
    assert X.shape[1] == 2**10
    _ = [dict(tt=x['text'], klass=x['klass']) for x in D]
    bow = BoW(lang='es', pretrain=False, key='tt',
              b4msa_kwargs=dict(max_dimension=True,
                                token_max_filter=2**10)).fit(_)
    X = bow.transform(_)
    assert X.shape[1] == 2**10
    _ = [dict(tt=x['text'], klass=x['klass'], tt2=x['text']) for x in D]
    bow = BoW(lang='es', pretrain=False, key=['tt', 'tt2'],
              b4msa_kwargs=dict(max_dimension=True,
                                token_max_filter=2**10)).fit(_)
    X = bow.transform(_)
    assert X.shape[1] == 2**10


def test_TextRepresentations_keyword():
    from EvoMSA.evodag import TextRepresentations
    text_repr = TextRepresentations(lang='es', keyword=True,
                                    emoji=False, dataset=False)
    X = text_repr.transform(['hola'])
    assert 2113 == X.shape[1]


def test_BoW_label_key():
    from EvoMSA.evodag import BoW
    D = list(tweet_iterator(TWEETS))
    _ = [dict(tt=x['text'], label=x['klass']) for x in D]
    bow = BoW(lang='es', pretrain=False, key='tt',
              label_key='label',  
              b4msa_kwargs=dict(max_dimension=True,
                                token_max_filter=2**10)).fit(_)
    assert bow.predict(['buenos días'])[0] == 'P'


def test_TextRepresentations_select2():
    from EvoMSA.evodag import TextRepresentations
    D = list(tweet_iterator(TWEETS))    
    text_repr = TextRepresentations(lang='es', 
                                    emoji=False,
                                    keyword=False,
                                    n_jobs=1)
    n_names = len(text_repr.names)
    text_repr.select(D=D)
    assert len(text_repr.names) < n_names


def test_BoW_names():
    from EvoMSA.evodag import BoW
    bow = BoW(lang='es')
    X = bow.transform(['hola'])
    assert len(bow.names) == X.shape[1]


def test_TextRepresentations_unit():
    from EvoMSA.evodag import TextRepresentations
    D = list(tweet_iterator(TWEETS))    
    text_repr = TextRepresentations(lang='es', 
                                    emoji=False,
                                    keyword=False,
                                    n_jobs=1,
                                    unit_vector=True)
    X = text_repr.transform(['buenos días', 'adios'])
    
    _ = np.sqrt((X ** 2).sum(axis=1))
    np.testing.assert_almost_equal(_, np.array([1, 1]))
    text_repr = TextRepresentations(lang='es', 
                                    emoji=False,
                                    keyword=False,
                                    n_jobs=1,
                                    key=['text', 'text'],
                                    unit_vector=True)
    X = text_repr.transform([dict(text='buenos días')])
    _ = np.sqrt((X ** 2).sum(axis=1))
    np.testing.assert_almost_equal(_, 1)